/* Copyright 2014-2018 The PySCF Developers. All Rights Reserved.
  
   Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at
 
        http://www.apache.org/licenses/LICENSE-2.0
 
    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.

 *
 * Author: Qiming Sun <osirpt.sun@gmail.com>
 */

void AO2MOnr_incore8f_acc(double *vout, double *eri, double *mo_coeff,
                          int row_start, int row_count, int nao,
                          int i_start, int i_count, int j_start, int j_count,
                          void (*ftrans)());
void AO2MOnr_incore4f_acc(double *vout, double *eri, double *mo_coeff,
                          int row_start, int row_count, int nao,
                          int i_start, int i_count, int j_start, int j_count,
                          void (*ftrans)());

void AO2MOnr_e1incore_drv(double *eri_mo, double *eri_ao, double *mo_coeff,
                          void (*facc)(), void (*ftrans)(), int nao,
                          int i_start, int i_count, int j_start, int j_count);

